/*======================================================================

    A general driver for accessing PCMCIA card memory via Bulk
    Memory Services.

    This driver provides the equivalent of /dev/mem for a PCMCIA
    card's attribute and common memory.  It includes character
    and block devices.

    Written by David Hinds, dhinds@allegro.stanford.edu
    
======================================================================*/


#include <linux/module.h>
#include <asm/uaccess.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/ptrace.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/major.h>
#include <linux/fs.h>
#include <linux/ioctl.h>
#include <asm/io.h>
#include <asm/system.h>
#include <asm/segment.h>
#include <stdarg.h>

#include "mtd.h"

struct mypriv {
	u_char *start;
	u_char *end;
};

void physmem_sync (struct mtd_info *mtd);
int physmem_erase (struct mtd_info *mtd, struct erase_info *instr);
int physmem_point (struct mtd_info *mtd, u_long from, u_long len, u_long *retlen, u_char **mtdbuf);
void physmem_unpoint (struct mtd_info *mtd, u_long addr);
int physmem_read (struct mtd_info *mtd, u_long from, u_long len, u_long *retlen, u_char *buf, int userspace);
int physmem_write (struct mtd_info *mtd, u_long to, u_long len, u_long *retlen, const u_char *buf, int userspace);
void physmem_openclose (struct mtd_info *mtd, int which);
	

void physmem_sync (struct mtd_info *mtd)
{
}

int physmem_erase (struct mtd_info *mtd, struct erase_info *instr)
{
	struct mypriv *priv = mtd->priv;

	if (instr->addr + instr->len > mtd->size)
		return -EINVAL;
	
	memset(priv->start + instr->addr, 0xff, instr->len);

	/* This'll catch a few races. Free the thing before returning :) 
	 * I don't feel at all ashamed. This kind of thing is possible anyway
	 * with flash, but unlikely.
	 */

	instr->state = MTD_ERASE_DONE;

	if (instr->callback)
		(*(instr->callback))(instr);
	else
		kfree(instr);

	return 0;
}


int physmem_point (struct mtd_info *mtd, u_long from, u_long len, u_long *retlen, u_char **mtdbuf)
{
	struct mypriv *priv = (struct mypriv *)mtd->priv;

	*mtdbuf = priv->start + from;
	*retlen = len;
	return 0;
}

void physmem_unpoint (struct mtd_info *mtd, u_long addr)
{
}

int physmem_read (struct mtd_info *mtd, u_long from, u_long len, u_long *retlen, u_char *buf, int userspace)
{
	struct mypriv *priv = (struct mypriv *)mtd->priv;
	
	*retlen=len;

	if (userspace)
		return copy_to_user(buf,priv->start+from, len);

	memcpy (buf, priv->start + from, len);
	return 0;
}

int physmem_write (struct mtd_info *mtd, u_long to, u_long len, u_long *retlen, const u_char *buf, int userspace)
{
	struct mypriv *priv = (struct mypriv *)mtd->priv;

	*retlen=len;
	if (userspace)
		return copy_from_user(priv->start+to, buf, len);

	memcpy (priv->start + to, buf, len);
	return 0;
}


void physmem_openclose (struct mtd_info *mtd, int which)
{
	if (which)
		MOD_INC_USE_COUNT;
	else
		MOD_DEC_USE_COUNT;
}


/*====================================================================*/

/* Place your defaults here */

static u_long start = 0;
static u_long length = 0;
static u_long end = 0;

MODULE_PARM(start,"l");
MODULE_PARM(length,"l");
MODULE_PARM(end,"l");

struct mtd_info *mymtd;

int init_module(void)
{
	if (!start)
	{
		printk(KERN_NOTICE "physmem: No start address for memory device.\n");
		return -EINVAL;
	}
	
	if (!length && !end)
	{
		printk(KERN_NOTICE "physmem: No length or endpointer given.\n");
		return -EINVAL;
	}
	
	if (!end)
		end = start + length;
	
	mymtd = kmalloc(sizeof(struct mtd_info), GFP_KERNEL);

	if (mymtd)
	{
		mymtd->priv = (void *) kmalloc (sizeof(struct mypriv), GFP_KERNEL);
		if (!mymtd->priv)
		{
			kfree(mymtd);
			mymtd = NULL;
		}
	}

	if (!mymtd)
	{
		printk(KERN_NOTICE "physmem: Cannot allocate new MTD device.\n");
		return -ENOMEM;
	}
	
	((struct mypriv *)mymtd->priv)->start = phys_to_virt(start);
	((struct mypriv *)mymtd->priv)->end = phys_to_virt(end);

	strcpy (&mymtd->name[0],"Raw memory");

	mymtd->size = length;
	mymtd->flags = MTD_CLEAR_BITS | MTD_SET_BITS | MTD_WRITEB_WRITEABLE | MTD_VOLATILE;
        mymtd->erase = physmem_erase;
	mymtd->point = physmem_point;
	mymtd->unpoint = physmem_unpoint;
	mymtd->read = physmem_read;
	mymtd->write = physmem_write;
	mymtd->openclose = physmem_openclose;
	mymtd->devtype = MTD_RAM;
	mymtd->erasesize = 0x10000;
	mymtd->devsize = length;

	if (add_mtd_device(mymtd))
	{
		printk("Failed to register new device\n");
		kfree(mymtd->priv);
		kfree(mymtd);
		return -EAGAIN;
	}

	return 0;
}

void cleanup_module(void)
{
	kfree (mymtd->priv);
	del_mtd_device(mymtd);
	kfree(mymtd);
}
